..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-emc_xtremio:

.. list-table:: Description of EMC XtremIO volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``xtremio_array_busy_retry_count`` = ``5``
     - (Integer) Number of retries in case array is busy
   * - ``xtremio_array_busy_retry_interval`` = ``5``
     - (Integer) Interval between retries in case array is busy
   * - ``xtremio_cluster_name`` =
     - (String) XMS cluster id in multi-cluster environment
   * - ``xtremio_volumes_per_glance_cache`` = ``100``
     - (Integer) Number of volumes created from each cached glance image
