..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-zfssa-iscsi:

.. list-table:: Description of ZFS Storage Appliance iSCSI driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``zfssa_initiator`` =
     - (String) iSCSI initiator IQNs. (comma separated)
   * - ``zfssa_initiator_config`` =
     - (String) iSCSI initiators configuration.
   * - ``zfssa_initiator_group`` =
     - (String) iSCSI initiator group.
   * - ``zfssa_initiator_password`` =
     - (String) Secret of the iSCSI initiator CHAP user.
   * - ``zfssa_initiator_user`` =
     - (String) iSCSI initiator CHAP user (name).
   * - ``zfssa_lun_compression`` = ``off``
     - (String) Data compression.
   * - ``zfssa_lun_logbias`` = ``latency``
     - (String) Synchronous write bias.
   * - ``zfssa_lun_sparse`` = ``False``
     - (Boolean) Flag to enable sparse (thin-provisioned): True, False.
   * - ``zfssa_lun_volblocksize`` = ``8k``
     - (String) Block size.
   * - ``zfssa_pool`` = ``None``
     - (String) Storage pool name.
   * - ``zfssa_project`` = ``None``
     - (String) Project name.
   * - ``zfssa_replication_ip`` =
     - (String) IP address used for replication data. (maybe the same as data ip)
   * - ``zfssa_rest_timeout`` = ``None``
     - (Integer) REST connection timeout. (seconds)
   * - ``zfssa_target_group`` = ``tgt-grp``
     - (String) iSCSI target group name.
   * - ``zfssa_target_interfaces`` = ``None``
     - (String) Network interfaces of iSCSI targets. (comma separated)
   * - ``zfssa_target_password`` =
     - (String) Secret of the iSCSI target CHAP user.
   * - ``zfssa_target_portal`` = ``None``
     - (String) iSCSI target portal (Data-IP:Port, w.x.y.z:3260).
   * - ``zfssa_target_user`` =
     - (String) iSCSI target CHAP user (name).
