..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-datera:

.. list-table:: Description of Datera volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``datera_503_interval`` = ``5``
     - (Integer) Interval between 503 retries
   * - ``datera_503_timeout`` = ``120``
     - (Integer) Timeout for HTTP 503 retry messages
   * - ``datera_api_port`` = ``7717``
     - (String) Datera API port.
   * - ``datera_api_version`` = ``2``
     - (String) DEPRECATED: Datera API version.
   * - ``datera_debug`` = ``False``
     - (Boolean) True to set function arg and return logging
   * - ``datera_debug_replica_count_override`` = ``False``
     - (Boolean) ONLY FOR DEBUG/TESTING PURPOSES True to set replica_count to 1
   * - ``datera_tenant_id`` = ``None``
     - (String) If set to 'Map' --> OpenStack project ID will be mapped implicitly to Datera tenant ID If set to 'None' --> Datera tenant ID will not be used during volume provisioning If set to anything else --> Datera tenant ID will be the provided value
