..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-hpmsa:

.. list-table:: Description of HP MSA volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``hpmsa_api_protocol`` = ``https``
     - (String) HPMSA API interface protocol.
   * - ``hpmsa_backend_name`` = ``A``
     - (String) Pool or Vdisk name to use for volume creation.
   * - ``hpmsa_backend_type`` = ``virtual``
     - (String) linear (for Vdisk) or virtual (for Pool).
   * - ``hpmsa_iscsi_ips`` =
     - (List) List of comma-separated target iSCSI IP addresses.
   * - ``hpmsa_verify_certificate`` = ``False``
     - (Boolean) Whether to verify HPMSA array SSL certificate.
   * - ``hpmsa_verify_certificate_path`` = ``None``
     - (String) HPMSA array SSL certificate path.
