..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-coprhd:

.. list-table:: Description of Coprhd volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``coprhd_emulate_snapshot`` = ``False``
     - (Boolean) True | False to indicate if the storage array in CoprHD is VMAX or VPLEX
   * - ``coprhd_hostname`` = ``None``
     - (String) Hostname for the CoprHD Instance
   * - ``coprhd_password`` = ``None``
     - (String) Password for accessing the CoprHD Instance
   * - ``coprhd_port`` = ``4443``
     - (Port number) Port for the CoprHD Instance
   * - ``coprhd_project`` = ``None``
     - (String) Project to utilize within the CoprHD Instance
   * - ``coprhd_scaleio_rest_gateway_host`` = ``None``
     - (String) Rest Gateway IP or FQDN for Scaleio
   * - ``coprhd_scaleio_rest_gateway_port`` = ``4984``
     - (Port number) Rest Gateway Port for Scaleio
   * - ``coprhd_scaleio_rest_server_password`` = ``None``
     - (String) Rest Gateway Password
   * - ``coprhd_scaleio_rest_server_username`` = ``None``
     - (String) Username for Rest Gateway
   * - ``coprhd_tenant`` = ``None``
     - (String) Tenant to utilize within the CoprHD Instance
   * - ``coprhd_username`` = ``None``
     - (String) Username for accessing the CoprHD Instance
   * - ``coprhd_varray`` = ``None``
     - (String) Virtual Array to utilize within the CoprHD Instance
   * - ``scaleio_server_certificate_path`` = ``None``
     - (String) Server certificate path
   * - ``scaleio_verify_server_certificate`` = ``False``
     - (Boolean) verify server certificate
