..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-backups_ceph:

.. list-table:: Description of Ceph backup driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``backup_ceph_chunk_size`` = ``134217728``
     - (Integer) The chunk size, in bytes, that a backup is broken into before transfer to the Ceph object store.
   * - ``backup_ceph_conf`` = ``/etc/ceph/ceph.conf``
     - (String) Ceph configuration file to use.
   * - ``backup_ceph_pool`` = ``backups``
     - (String) The Ceph pool where volume backups are stored.
   * - ``backup_ceph_stripe_count`` = ``0``
     - (Integer) RBD stripe count to use when creating a backup image.
   * - ``backup_ceph_stripe_unit`` = ``0``
     - (Integer) RBD stripe unit to use when creating a backup image.
   * - ``backup_ceph_user`` = ``cinder``
     - (String) The Ceph user to connect with. Default here is to use the same user as for Cinder volumes. If not using cephx this should be set to None.
   * - ``restore_discard_excess_bytes`` = ``True``
     - (Boolean) If True, always discard excess bytes when restoring volumes i.e. pad with zeroes.
