..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-synology:

.. list-table:: Description of Synology volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``synology_admin_port`` = ``5000``
     - (Port number) Management port for Synology storage.
   * - ``synology_device_id`` = ``None``
     - (String) Device id for skip one time password check for logging in Synology storage if OTP is enabled.
   * - ``synology_one_time_pass`` = ``None``
     - (String) One time password of administrator for logging in Synology storage if OTP is enabled.
   * - ``synology_password`` =
     - (String) Password of administrator for logging in Synology storage.
   * - ``synology_pool_name`` =
     - (String) Volume on Synology storage to be used for creating lun.
   * - ``synology_ssl_verify`` = ``True``
     - (Boolean) Do certificate validation or not if $driver_use_ssl is True
   * - ``synology_username`` = ``admin``
     - (String) Administrator of Synology storage.
