..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-zfssa-nfs:

.. list-table:: Description of ZFS Storage Appliance NFS driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``zfssa_cache_directory`` = ``os-cinder-cache``
     - (String) Name of directory inside zfssa_nfs_share where cache volumes are stored.
   * - ``zfssa_cache_project`` = ``os-cinder-cache``
     - (String) Name of ZFSSA project where cache volumes are stored.
   * - ``zfssa_data_ip`` = ``None``
     - (String) Data path IP address
   * - ``zfssa_enable_local_cache`` = ``True``
     - (Boolean) Flag to enable local caching: True, False.
   * - ``zfssa_https_port`` = ``443``
     - (String) HTTPS port number
   * - ``zfssa_manage_policy`` = ``loose``
     - (String) Driver policy for volume manage.
   * - ``zfssa_nfs_mount_options`` =
     - (String) Options to be passed while mounting share over nfs
   * - ``zfssa_nfs_pool`` =
     - (String) Storage pool name.
   * - ``zfssa_nfs_project`` = ``NFSProject``
     - (String) Project name.
   * - ``zfssa_nfs_share`` = ``nfs_share``
     - (String) Share name.
   * - ``zfssa_nfs_share_compression`` = ``off``
     - (String) Data compression.
   * - ``zfssa_nfs_share_logbias`` = ``latency``
     - (String) Synchronous write bias-latency, throughput.
   * - ``zfssa_rest_timeout`` = ``None``
     - (Integer) REST connection timeout. (seconds)
