..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-common:

.. list-table:: Description of common configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``allow_availability_zone_fallback`` = ``False``
     - (Boolean) If the requested Cinder availability zone is unavailable, fall back to the value of default_availability_zone, then storage_availability_zone, instead of failing.
   * - ``chap`` = ``disabled``
     - (String) CHAP authentication mode, effective only for iscsi (disabled|enabled)
   * - ``chap_password`` =
     - (String) Password for specified CHAP account name.
   * - ``chap_username`` =
     - (String) CHAP user name.
   * - ``chiscsi_conf`` = ``/etc/chelsio-iscsi/chiscsi.conf``
     - (String) Chiscsi (CXT) global defaults configuration file
   * - ``cinder_internal_tenant_project_id`` = ``None``
     - (String) ID of the project which will be used as the Cinder internal tenant.
   * - ``cinder_internal_tenant_user_id`` = ``None``
     - (String) ID of the user to be used in volume operations as the Cinder internal tenant.
   * - ``cluster`` = ``None``
     - (String) Name of this cluster. Used to group volume hosts that share the same backend configurations to work in HA Active-Active mode. Active-Active is not yet supported.
   * - ``compute_api_class`` = ``cinder.compute.nova.API``
     - (String) The full class name of the compute API class to use
   * - ``connection_type`` = ``iscsi``
     - (String) Connection type to the IBM Storage Array
   * - ``consistencygroup_api_class`` = ``cinder.consistencygroup.api.API``
     - (String) The full class name of the consistencygroup API class
   * - ``default_availability_zone`` = ``None``
     - (String) Default availability zone for new volumes. If not set, the storage_availability_zone option value is used as the default for new volumes.
   * - ``default_group_type`` = ``None``
     - (String) Default group type to use
   * - ``default_volume_type`` = ``None``
     - (String) Default volume type to use
   * - ``driver_client_cert`` = ``None``
     - (String) The path to the client certificate for verification, if the driver supports it.
   * - ``driver_client_cert_key`` = ``None``
     - (String) The path to the client certificate key for verification, if the driver supports it.
   * - ``driver_data_namespace`` = ``None``
     - (String) Namespace for driver private data values to be saved in.
   * - ``driver_ssl_cert_path`` = ``None``
     - (String) Can be used to specify a non default path to a CA_BUNDLE file or directory with certificates of trusted CAs, which will be used to validate the backend
   * - ``driver_ssl_cert_verify`` = ``False``
     - (Boolean) If set to True the http client will validate the SSL certificate of the backend endpoint.
   * - ``enable_force_upload`` = ``False``
     - (Boolean) Enables the Force option on upload_to_image. This enables running upload_volume on in-use volumes for backends that support it.
   * - ``enable_new_services`` = ``True``
     - (Boolean) Services to be added to the available pool on create
   * - ``enable_unsupported_driver`` = ``False``
     - (Boolean) Set this to True when you want to allow an unsupported driver to start. Drivers that haven't maintained a working CI system and testing are marked as unsupported until CI is working again. This also marks a driver as deprecated and may be removed in the next release.
   * - ``end_time`` = ``None``
     - (String) If this option is specified then the end time specified is used instead of the end time of the last completed audit period.
   * - ``enforce_multipath_for_image_xfer`` = ``False``
     - (Boolean) If this is set to True, attachment of volumes for image transfer will be aborted when multipathd is not running. Otherwise, it will fallback to single path.
   * - ``executor_thread_pool_size`` = ``64``
     - (Integer) Size of executor thread pool.
   * - ``fatal_exception_format_errors`` = ``False``
     - (Boolean) Make exception message format errors fatal.
   * - ``group_api_class`` = ``cinder.group.api.API``
     - (String) The full class name of the group API class
   * - ``host`` = ``localhost``
     - (String) Name of this node. This can be an opaque identifier. It is not necessarily a host name, FQDN, or IP address.
   * - ``iet_conf`` = ``/etc/iet/ietd.conf``
     - (String) IET configuration file
   * - ``iscsi_secondary_ip_addresses`` =
     - (List) The list of secondary IP addresses of the iSCSI daemon
   * - ``max_over_subscription_ratio`` = ``20.0``
     - (Floating point) Float representation of the over subscription ratio when thin provisioning is involved. Default ratio is 20.0, meaning provisioned capacity can be 20 times of the total physical capacity. If the ratio is 10.5, it means provisioned capacity can be 10.5 times of the total physical capacity. A ratio of 1.0 means provisioned capacity cannot exceed the total physical capacity. The ratio has to be a minimum of 1.0.
   * - ``monkey_patch`` = ``False``
     - (Boolean) Enable monkey patching
   * - ``monkey_patch_modules`` =
     - (List) List of modules/decorators to monkey patch
   * - ``my_ip`` = ``10.0.0.1``
     - (String) IP address of this host
   * - ``no_snapshot_gb_quota`` = ``False``
     - (Boolean) Whether snapshots count against gigabyte quota
   * - ``num_shell_tries`` = ``3``
     - (Integer) Number of times to attempt to run flakey shell commands
   * - ``os_privileged_user_auth_url`` = ``None``
     - (URI) Auth URL associated with the OpenStack privileged account.
   * - ``os_privileged_user_name`` = ``None``
     - (String) OpenStack privileged account username. Used for requests to other services (such as Nova) that require an account with special rights.
   * - ``os_privileged_user_password`` = ``None``
     - (String) Password associated with the OpenStack privileged account.
   * - ``os_privileged_user_tenant`` = ``None``
     - (String) Tenant name associated with the OpenStack privileged account.
   * - ``periodic_fuzzy_delay`` = ``60``
     - (Integer) Range, in seconds, to randomly delay when starting the periodic task scheduler to reduce stampeding. (Disable by setting to 0)
   * - ``periodic_interval`` = ``60``
     - (Integer) Interval, in seconds, between running periodic tasks
   * - ``replication_device`` = ``None``
     - (Unknown) Multi opt of dictionaries to represent a replication target device. This option may be specified multiple times in a single config section to specify multiple replication target devices. Each entry takes the standard dict config form: replication_device = target_device_id:<required>,key1:value1,key2:value2...
   * - ``report_discard_supported`` = ``False``
     - (Boolean) Report to clients of Cinder that the backend supports discard (aka. trim/unmap). This will not actually change the behavior of the backend or the client directly, it will only notify that it can be used.
   * - ``report_interval`` = ``10``
     - (Integer) Interval, in seconds, between nodes reporting state to datastore
   * - ``reserved_percentage`` = ``0``
     - (Integer) The percentage of backend capacity is reserved
   * - ``rootwrap_config`` = ``/etc/cinder/rootwrap.conf``
     - (String) Path to the rootwrap configuration file to use for running commands as root
   * - ``send_actions`` = ``False``
     - (Boolean) Send the volume and snapshot create and delete notifications generated in the specified period.
   * - ``service_down_time`` = ``60``
     - (Integer) Maximum time since last check-in for a service to be considered up
   * - ``ssh_hosts_key_file`` = ``$state_path/ssh_known_hosts``
     - (String) File containing SSH host keys for the systems with which Cinder needs to communicate. OPTIONAL: Default=$state_path/ssh_known_hosts
   * - ``start_time`` = ``None``
     - (String) If this option is specified then the start time specified is used instead of the start time of the last completed audit period.
   * - ``state_path`` = ``/var/lib/cinder``
     - (String) Top-level directory for maintaining cinder's state
   * - ``storage_availability_zone`` = ``nova``
     - (String) Availability zone of this node
   * - ``storage_protocol`` = ``iscsi``
     - (String) Protocol for transferring data between host and storage back-end.
   * - ``strict_ssh_host_key_policy`` = ``False``
     - (Boolean) Option to enable strict host key checking. When set to "True" Cinder will only connect to systems with a host key present in the configured "ssh_hosts_key_file". When set to "False" the host key will be saved upon first connection and used for subsequent connections. Default=False
   * - ``suppress_requests_ssl_warnings`` = ``False``
     - (Boolean) Suppress requests library SSL certificate warnings.
   * - ``tcp_keepalive`` = ``True``
     - (Boolean) Sets the value of TCP_KEEPALIVE (True/False) for each server socket.
   * - ``tcp_keepalive_count`` = ``None``
     - (Integer) Sets the value of TCP_KEEPCNT for each server socket. Not supported on OS X.
   * - ``tcp_keepalive_interval`` = ``None``
     - (Integer) Sets the value of TCP_KEEPINTVL in seconds for each server socket. Not supported on OS X.
   * - ``until_refresh`` = ``0``
     - (Integer) Count of reservations until usage is refreshed
   * - ``use_chap_auth`` = ``False``
     - (Boolean) Option to enable/disable CHAP authentication for targets.
   * - ``use_forwarded_for`` = ``False``
     - (Boolean) Treat X-Forwarded-For as the canonical remote address. Only enable this if you have a sanitizing proxy.
   * - **[healthcheck]**
     -
   * - ``backends`` =
     - (List) Additional backends that can perform health checks and report that information back as part of a request.
   * - ``detailed`` = ``False``
     - (Boolean) Show more detailed information as part of the response
   * - ``disable_by_file_path`` = ``None``
     - (String) Check the presence of a file to determine if an application is running on a port. Used by DisableByFileHealthcheck plugin.
   * - ``disable_by_file_paths`` =
     - (List) Check the presence of a file based on a port to determine if an application is running on a port. Expects a "port:path" list of strings. Used by DisableByFilesPortsHealthcheck plugin.
   * - ``path`` = ``/healthcheck``
     - (String) DEPRECATED: The path to respond to healtcheck requests on.
   * - **[key_manager]**
     -
   * - ``backend`` = ``barbican``
     - (String) The full class name of the key manager API class
   * - ``fixed_key`` = ``None``
     - (String) Fixed key returned by key manager, specified in hex
