..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-backups_gcs:

.. list-table:: Description of GCS backup driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``backup_gcs_block_size`` = ``32768``
     - (Integer) The size in bytes that changes are tracked for incremental backups. backup_gcs_object_size has to be multiple of backup_gcs_block_size.
   * - ``backup_gcs_bucket`` = ``None``
     - (String) The GCS bucket to use.
   * - ``backup_gcs_bucket_location`` = ``US``
     - (String) Location of GCS bucket.
   * - ``backup_gcs_credential_file`` = ``None``
     - (String) Absolute path of GCS service account credential file.
   * - ``backup_gcs_enable_progress_timer`` = ``True``
     - (Boolean) Enable or Disable the timer to send the periodic progress notifications to Ceilometer when backing up the volume to the GCS backend storage. The default value is True to enable the timer.
   * - ``backup_gcs_num_retries`` = ``3``
     - (Integer) Number of times to retry.
   * - ``backup_gcs_object_size`` = ``52428800``
     - (Integer) The size in bytes of GCS backup objects.
   * - ``backup_gcs_project_id`` = ``None``
     - (String) Owner project id for GCS bucket.
   * - ``backup_gcs_proxy_url`` = ``None``
     - (URI) URL for http proxy access.
   * - ``backup_gcs_reader_chunk_size`` = ``2097152``
     - (Integer) GCS object will be downloaded in chunks of bytes.
   * - ``backup_gcs_retry_error_codes`` = ``429``
     - (List) List of GCS error codes.
   * - ``backup_gcs_storage_class`` = ``NEARLINE``
     - (String) Storage class of GCS bucket.
   * - ``backup_gcs_user_agent`` = ``gcscinder``
     - (String) Http user-agent string for gcs api.
   * - ``backup_gcs_writer_chunk_size`` = ``2097152``
     - (Integer) GCS object will be uploaded in chunks of bytes. Pass in a value of -1 if the file is to be uploaded as a single chunk.
