..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-images:

.. list-table:: Description of images configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``allowed_direct_url_schemes`` =
     - (List) A list of url schemes that can be downloaded directly via the direct_url. Currently supported schemes: [file, cinder].
   * - ``glance_api_insecure`` = ``False``
     - (Boolean) Allow to perform insecure SSL (https) requests to glance (https will be used but cert validation will not be performed).
   * - ``glance_api_servers`` = ``None``
     - (List) A list of the URLs of glance API servers available to cinder ([http[s]://][hostname|ip]:port). If protocol is not specified it defaults to http.
   * - ``glance_api_ssl_compression`` = ``False``
     - (Boolean) Enables or disables negotiation of SSL layer compression. In some cases disabling compression can improve data throughput, such as when high network bandwidth is available and you use compressed image formats like qcow2.
   * - ``glance_ca_certificates_file`` = ``None``
     - (String) Location of ca certificates file to use for glance client requests.
   * - ``glance_catalog_info`` = ``image:glance:publicURL``
     - (String) Info to match when looking for glance in the service catalog. Format is: separated values of the form: <service_type>:<service_name>:<endpoint_type> - Only used if glance_api_servers are not provided.
   * - ``glance_core_properties`` = ``checksum, container_format, disk_format, image_name, image_id, min_disk, min_ram, name, size``
     - (List) Default core properties of image
   * - ``glance_num_retries`` = ``0``
     - (Integer) Number retries when downloading an image from glance
   * - ``glance_request_timeout`` = ``None``
     - (Integer) http/https timeout value for glance operations. If no value (None) is supplied here, the glanceclient default value is used.
   * - ``image_conversion_dir`` = ``$state_path/conversion``
     - (String) Directory used for temporary storage during image conversion
   * - ``image_upload_use_cinder_backend`` = ``False``
     - (Boolean) If set to True, upload-to-image in raw format will create a cloned volume and register its location to the image service, instead of uploading the volume content. The cinder backend and locations support must be enabled in the image service.
   * - ``image_upload_use_internal_tenant`` = ``False``
     - (Boolean) If set to True, the image volume created by upload-to-image will be placed in the internal tenant. Otherwise, the image volume is created in the current context's tenant.
   * - ``image_volume_cache_enabled`` = ``False``
     - (Boolean) Enable the image volume cache for this backend.
   * - ``image_volume_cache_max_count`` = ``0``
     - (Integer) Max number of entries allowed in the image volume cache. 0 => unlimited.
   * - ``image_volume_cache_max_size_gb`` = ``0``
     - (Integer) Max size of the image volume cache for this backend in GB. 0 => unlimited.
   * - ``use_multipath_for_image_xfer`` = ``False``
     - (Boolean) Do we attach/detach volumes in cinder using multipath for volume to image and image to volume transfers?
