..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-eqlx:

.. list-table:: Description of Dell EqualLogic volume driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``eqlx_cli_max_retries`` = ``5``
     - (Integer) Maximum retry count for reconnection. Default is 5.
   * - ``eqlx_group_name`` = ``group-0``
     - (String) Group name to use for creating volumes. Defaults to "group-0".
   * - ``eqlx_pool`` = ``default``
     - (String) Pool in which volumes will be created. Defaults to "default".
