..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-backups_swift:

.. list-table:: Description of Swift backup driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``backup_swift_auth`` = ``per_user``
     - (String) Swift authentication mechanism
   * - ``backup_swift_auth_version`` = ``1``
     - (String) Swift authentication version. Specify "1" for auth 1.0, or "2" for auth 2.0 or "3" for auth 3.0
   * - ``backup_swift_block_size`` = ``32768``
     - (Integer) The size in bytes that changes are tracked for incremental backups. backup_swift_object_size has to be multiple of backup_swift_block_size.
   * - ``backup_swift_ca_cert_file`` = ``None``
     - (String) Location of the CA certificate file to use for swift client requests.
   * - ``backup_swift_container`` = ``volumebackups``
     - (String) The default Swift container to use
   * - ``backup_swift_enable_progress_timer`` = ``True``
     - (Boolean) Enable or Disable the timer to send the periodic progress notifications to Ceilometer when backing up the volume to the Swift backend storage. The default value is True to enable the timer.
   * - ``backup_swift_key`` = ``None``
     - (String) Swift key for authentication
   * - ``backup_swift_object_size`` = ``52428800``
     - (Integer) The size in bytes of Swift backup objects
   * - ``backup_swift_project`` = ``None``
     - (String) Swift project/account name. Required when connecting to an auth 3.0 system
   * - ``backup_swift_project_domain`` = ``None``
     - (String) Swift project domain name. Required when connecting to an auth 3.0 system
   * - ``backup_swift_retry_attempts`` = ``3``
     - (Integer) The number of retries to make for Swift operations
   * - ``backup_swift_retry_backoff`` = ``2``
     - (Integer) The backoff time in seconds between Swift retries
   * - ``backup_swift_tenant`` = ``None``
     - (String) Swift tenant/account name. Required when connecting to an auth 2.0 system
   * - ``backup_swift_url`` = ``None``
     - (URI) The URL of the Swift endpoint
   * - ``backup_swift_user`` = ``None``
     - (String) Swift user name
   * - ``backup_swift_user_domain`` = ``None``
     - (String) Swift user domain name. Required when connecting to an auth 3.0 system
   * - ``keystone_catalog_info`` = ``identity:Identity Service:publicURL``
     - (String) Info to match when looking for keystone in the service catalog. Format is: separated values of the form: <service_type>:<service_name>:<endpoint_type> - Only used if backup_swift_auth_url is unset
   * - ``swift_catalog_info`` = ``object-store:swift:publicURL``
     - (String) Info to match when looking for swift in the service catalog. Format is: separated values of the form: <service_type>:<service_name>:<endpoint_type> - Only used if backup_swift_url is unset
