..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-san:

.. list-table:: Description of SAN configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``san_clustername`` =
     - (String) Cluster name to use for creating volumes
   * - ``san_ip`` =
     - (String) IP address of SAN controller
   * - ``san_is_local`` = ``False``
     - (Boolean) Execute commands locally instead of over SSH; use if the volume service is running on the SAN device
   * - ``san_login`` = ``admin``
     - (String) Username for SAN controller
   * - ``san_password`` =
     - (String) Password for SAN controller
   * - ``san_private_key`` =
     - (String) Filename of private key to use for SSH authentication
   * - ``san_ssh_port`` = ``22``
     - (Port number) SSH port to use with SAN
   * - ``san_thin_provision`` = ``True``
     - (Boolean) Use thin provisioning for SAN volumes?
   * - ``ssh_conn_timeout`` = ``30``
     - (Integer) SSH connection timeout in seconds
   * - ``ssh_max_pool_conn`` = ``5``
     - (Integer) Maximum ssh connections in the pool
   * - ``ssh_min_pool_conn`` = ``1``
     - (Integer) Minimum ssh connections in the pool
