..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-drbd:

.. list-table:: Description of DRBD configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``drbdmanage_devs_on_controller`` = ``True``
     - (Boolean) If set, the c-vol node will receive a useable /dev/drbdX device, even if the actual data is stored on other nodes only. This is useful for debugging, maintenance, and to be able to do the iSCSI export from the c-vol node.
   * - ``drbdmanage_disk_options`` = ``{"c-min-rate": "4M"}``
     - (String) Disk options to set on new resources. See http://www.drbd.org/en/doc/users-guide-90/re-drbdconf for all the details.
   * - ``drbdmanage_net_options`` = ``{"connect-int": "4", "allow-two-primaries": "yes", "ko-count": "30", "max-buffers": "20000", "ping-timeout": "100"}``
     - (String) Net options to set on new resources. See http://www.drbd.org/en/doc/users-guide-90/re-drbdconf for all the details.
   * - ``drbdmanage_redundancy`` = ``1``
     - (Integer) Number of nodes that should replicate the data.
   * - ``drbdmanage_resize_plugin`` = ``drbdmanage.plugins.plugins.wait_for.WaitForVolumeSize``
     - (String) Volume resize completion wait plugin.
   * - ``drbdmanage_resize_policy`` = ``{"timeout": "60"}``
     - (String) Volume resize completion wait policy.
   * - ``drbdmanage_resource_options`` = ``{"auto-promote-timeout": "300"}``
     - (String) Resource options to set on new resources. See http://www.drbd.org/en/doc/users-guide-90/re-drbdconf for all the details.
   * - ``drbdmanage_resource_plugin`` = ``drbdmanage.plugins.plugins.wait_for.WaitForResource``
     - (String) Resource deployment completion wait plugin.
   * - ``drbdmanage_resource_policy`` = ``{"ratio": "0.51", "timeout": "60"}``
     - (String) Resource deployment completion wait policy.
   * - ``drbdmanage_snapshot_plugin`` = ``drbdmanage.plugins.plugins.wait_for.WaitForSnapshot``
     - (String) Snapshot completion wait plugin.
   * - ``drbdmanage_snapshot_policy`` = ``{"count": "1", "timeout": "60"}``
     - (String) Snapshot completion wait policy.
