..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-zadara:

.. list-table:: Description of Zadara Storage driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``zadara_default_snap_policy`` = ``False``
     - (Boolean) VPSA - Attach snapshot policy for volumes
   * - ``zadara_password`` = ``None``
     - (String) VPSA - Password
   * - ``zadara_use_iser`` = ``True``
     - (Boolean) VPSA - Use ISER instead of iSCSI
   * - ``zadara_user`` = ``None``
     - (String) VPSA - Username
   * - ``zadara_vol_encrypt`` = ``False``
     - (Boolean) VPSA - Default encryption policy for volumes
   * - ``zadara_vol_name_template`` = ``OS_%s``
     - (String) VPSA - Default template for VPSA volume names
   * - ``zadara_vpsa_host`` = ``None``
     - (String) VPSA - Management Host name or IP address
   * - ``zadara_vpsa_poolname`` = ``None``
     - (String) VPSA - Storage Pool assigned for volumes
   * - ``zadara_vpsa_port`` = ``None``
     - (Port number) VPSA - Port number
   * - ``zadara_vpsa_use_ssl`` = ``False``
     - (Boolean) VPSA - Use SSL connection
